////////////////////////////////////////
//(c) 2010 EXAR Corporation
//All Rights Reserved.
//
//Exar Corporation ("Exar") hereby grants the User of this Exar
//product-specific XR17C15x, xr17v25x, and xr17v35x Windows CE 6.0 Device Driver (hereinafter
//referred to as "Driver") a non-exclusive, nontransferable license
//to use the Driver in accordance with the term set forth below.
//Before using the Driver, the User should read the following use
//restrictions. If the User does not accept these terms, the Driver
//should not be used or downloaded.
//The User is granted this limited license to use this Driver for
//User's end product that includes the Exar XR17C158/XR17D158/XR17V258/XR17V358 Octal UART PCI
//device, XR17C154/XR17D154/XR17V254/XR17V354 Quad UART PCI Device, and XR17C152/XR17D152/XR17V252/XR17V352
//Dual UART PCI Device and is not granted rights to sell, loan, rent, lease or
//license the Driver, in whole or in part, or in modified form to anyone
//other than User. User may modify the Driver to suit its specific
//applications but rights to derivative works and such modifications
//shall belong to Exar.
//
//The Driver is provided on an "AS IS" basis and Exar makes absolutely
//no warranty with respect to performance or the information contained
//therein. EXAR DISCLAIMS AND USER WAIVES ALL WARRANTIES, EXPRESS OR
//IMPLIED, INCLUDING WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
//PARTICULAR PURPOSE. The entire risk as to Driver quality and
//performance is with the User. ACCORDINGLY, IN NO EVENT SHALL EXAR
//BE LIABLE FOR ANY DAMAGES, WHETHER IN CONTRACT OR TORT, INCLUDING
//ANY LOST PROFITS OR OTHER INCIDENTAL, CONSEQUENTIAL, EXEMPLARY, OR
//PUNITIVE DAMAGES ARISING OUT OF THE USE OR APPLICATION OF THE DRIVER.
//Further, Exar reserves the right to make changes tot eh Driver without
//notice to improve reliability, function or design. Exar does not
//convey any license under patent rights or any other intellectual
//property rights, including those of third parties.
//
//Exar is not obligated to provide maintenance or support for the Driver.
//
////////////////////////////////////////////////////////
/*++

Module Name:

	xrpci.c

Abstract:

	Exar PCI UART card enumerator device driver

Notes:

	This driver loads to any of the Exar PCI cards (0x158/154/152/0x258/0x254/0x252 PIDs) and
	enumerates each port on the card with the proper I/O and interrupt
	resources. This can be considered as a Bus Driver for the card.

--*/

#include <windows.h>
#include <pkfuncs.h>
#include <ceddk.h>
#include <devload.h>
#include <pm.h>
#include <nkintr.h>
#include <winioctl.h>
#include <ddkreg.h>

#include "xrpci.h"

#ifdef DEBUG

#define DEBUGMASK(bit)      (1 << (bit))

#define MASK_ERROR          DEBUGMASK(0)
#define MASK_WARN           DEBUGMASK(1)
#define MASK_INIT           DEBUGMASK(2)
#define MASK_FUNCTION       DEBUGMASK(3)
#define MASK_IOCTL          DEBUGMASK(4)
#define MASK_DEVICE         DEBUGMASK(5)
#define MASK_ACTIVITY       DEBUGMASK(6)

DBGPARAM dpCurSettings = {
	_T("PMSampleDev"),
	{
		_T("Errors"), _T("Warnings"), _T("Init"), _T("Function"),
		_T("Ioctl"), _T("Device"), _T("Activity"), _T(""),
		_T(""),_T(""),_T(""),_T(""),
		_T(""),_T(""),_T(""),_T("")
	},
	MASK_ERROR | MASK_WARN | MASK_INIT | MASK_IOCTL | MASK_DEVICE
};

#define ZONE_ERROR          DEBUGZONE(0)
#define ZONE_WARN           DEBUGZONE(1)
#define ZONE_INIT           DEBUGZONE(2)
#define ZONE_FUNCTION       DEBUGZONE(3)
#define ZONE_IOCTL          DEBUGZONE(4)
#define ZONE_DEVICE         DEBUGZONE(5)
#define ZONE_ACTIVITY       DEBUGZONE(6)

#endif


DWORD
XRP_Init(
	PVOID Context
	)
{
	DWORD dwStatus, dwType, dwSize;
	HKEY hkDevice, hk;
	TCHAR szName[16];
	UCHAR i, dwPorts;
	DWORD dwDisp, dwIOBase, dwSysIntr, dwPortInstanceOnChip;
	WCHAR szLine[MAX_PATH];
	DWORD dwIOLength, dwIndex, dwIoctl = 4;
	DWORD dwRegStride;

	// unimodem
	DWORD dwDeviceType;
	UCHAR ucDevConfig[] = {0x10,0x00, 0x00,0x00, 0x05,0x00,0x00,0x00, 0x10,0x01,0x00,0x00, 0x00,0x4B,0x00,0x00, 0x00,0x00, 0x08, 0x00, 0x00, 0x00,0x00,0x00,0x00};

                                        //@@@ Debug Set to 0
    static DWORD dwDeviceArrayIndex = 1; //because COM1 with DeviceArrayIndex=0 is already assigned to Serial0 in our platform.reg
	static DWORD dwCardInstance = 0;
    DWORD robVer=0;

	PXRPCI_INFO pHWHead;

	UNREFERENCED_PARAMETER(Context);

    RETAILMSG( 1, (_T("XRP_Init: context %s\n"), Context));

	// get our activation information
	dwStatus = RegOpenKeyEx(HKEY_LOCAL_MACHINE, (LPCTSTR) Context, 0, 0, &hkDevice);
	if(dwStatus != ERROR_SUCCESS)
	{
        RETAILMSG( 1, (_T("XRP_Init: OpenDeviceKey('%s') failed %u\r\n"), Context, dwStatus));
		return 0;
	}

	dwSize = sizeof(szName);
	dwStatus = RegQueryValueEx(hkDevice, DEVLOAD_DEVNAME_VALNAME, NULL, &dwType, (LPBYTE) szName, &dwSize);
	if(dwStatus != ERROR_SUCCESS || dwType != DEVLOAD_DEVNAME_VALTYPE)
	{
        RETAILMSG( 1, (_T("XRP_Init: RegQueryValueEx('%s', '%s') failed %u\r\n"),
			Context, DEVLOAD_DEVNAME_VALNAME, dwStatus));
		RegCloseKey(hkDevice);
		return 0;
	}
    RETAILMSG( 1, (_T("XRP_Init: device name is '%s'\r\n"), szName));
	RegCloseKey(hkDevice);

	// Get the PCI data

	// Allocate for our main data structure and one of it's fields.
	pHWHead = (PXRPCI_INFO)LocalAlloc( LMEM_ZEROINIT|LMEM_FIXED , sizeof(XRPCI_INFO) );
	if ( !pHWHead )
		return( 0 );

	if ( ! XRPCI_GetRegistryData(pHWHead, (LPCTSTR)Context) )
	{
        RETAILMSG( 1, (TEXT("XRPCI_GetRegistryData - Unable to read registry data. Failing Init !!! \r\n")));

		wsprintf (szLine, L"Drivers\\BuiltIn\\PCI\\Template\\PureDebug");
		if (ERROR_SUCCESS == RegCreateKeyEx (HKEY_LOCAL_MACHINE, szLine, 0, NULL, 0, KEY_ALL_ACCESS, NULL, &hk, &dwDisp))
		{
			RegSetValueEx (hk, L"Blame", 0, REG_SZ, (BYTE *)L"XRPCI_GetRegistryData", sizeof(L"XRPCI_GetRegistryData"));
			RegCloseKey (hk);
		}

		return 0;
	}

	//
	// populate the ports
	//
	dwPorts = 4;
	dwIOLength = 0x8;

	//
	// update Exar's PCI device instance
	//
	dwCardInstance++;


	//
	// UART register stride (default = 1 byte).
	//
	dwRegStride = 1;

    RETAILMSG( 1, (_T("dwPorts = %d \r\n"), dwPorts));

	for (i = 0; i < dwPorts; i++)
	{
		//
		//we are ready for loading drivers for the ports, so update the registry
		//
		dwIOBase = pHWHead->dwIOBase + (i * dwIOLength);
		if( !KernelIoControl(IOCTL_HAL_REQUEST_SYSINTR, &pHWHead->dwIrq, sizeof(pHWHead->dwIrq), &dwSysIntr, sizeof(dwSysIntr), NULL) )
		{
			//error just use the default System Interrupt ; This should not happen though...
			dwSysIntr = pHWHead->dwSysIntr;
		}

		//
		// The registry name for individual ports would be
		// (\xrserial"cardinstance""deviceid""portindex")
		// ex: xrserial11580 - would be card "1", which is "0x158" device, port "0"
		// xrserial21521 - would be card "2", which is "0x152" device, port "1"
		// dwCardInstance starts with index 1 and port index starts with 0
		// This way we can support any number of cards
		//
        //wsprintf (szLine, L"Drivers\\BuiltIn\\PCI\\Template\\xrserial%d%x%d", dwCardInstance, pHWHead->dwDeviceId, i);
        wsprintf (szLine, L"Drivers\\xrserial%d%x%d", dwCardInstance, pHWHead->dwDeviceId, i);
		if (ERROR_SUCCESS == RegCreateKeyEx (HKEY_LOCAL_MACHINE, szLine, 0, NULL, 0, KEY_ALL_ACCESS, NULL, &hk, &dwDisp))
		{
            robVer = 3;
            RegSetValueEx (hk, L"Rob", 0, REG_DWORD, (BYTE *)&robVer, sizeof(DWORD));
            RegSetValueEx (hk, L"SysIntr", 0, REG_DWORD, (BYTE *)&dwSysIntr, sizeof(DWORD));
			RegSetValueEx (hk, L"IRQ", 0, REG_DWORD, (BYTE *)&pHWHead->dwIrq, sizeof(DWORD));
			RegSetValueEx (hk, L"DevIndex", 0, REG_DWORD, (BYTE *)&pHWHead->dwDevIndex, sizeof(DWORD));
			RegSetValueEx (hk, L"InterfaceType", 0, REG_DWORD, (BYTE *)&pHWHead->dwInterfaceType, sizeof(DWORD));
			RegSetValueEx (hk, L"BusNumber", 0, REG_DWORD, (BYTE *)&pHWHead->dwBusNumber, sizeof(DWORD));
			RegSetValueEx (hk, L"IoBase", 0, REG_DWORD, (BYTE *)&dwIOBase, sizeof(DWORD));
            RegSetValueEx (hk, L"IoLen", 0, REG_DWORD, (BYTE*)&dwIOLength, sizeof(DWORD));
			RegSetValueEx (hk, L"RegStride", 0, REG_DWORD, (BYTE *)&dwRegStride, sizeof(DWORD));
			//RegSetValueEx (hk, L"IsrDll", 0, REG_SZ, (BYTE *)L"xrisr.dll", sizeof(L"xrisr.dll"));
			//RegSetValueEx (hk, L"IsrHandler", 0, REG_SZ, (BYTE *)L"ISRHandler", sizeof(L"ISRHandler"));
            //RegSetValueEx (hk, L"Dll", 0, REG_SZ, (BYTE *)L"com16550.dll", sizeof(L"com16550.dll"));
            //RegSetValueEx (hk, L"IsrDll", 0, REG_SZ, (BYTE *)L"isr16550.dll", sizeof(L"isr16550.dll"));
            RegSetValueEx (hk, L"Dll", 0, REG_SZ, (BYTE *)L"\\Hard Disk\\AIO_COM\\comaccesio.dll", sizeof(L"\\Hard Disk\\AIO_COM\\comaccesio.dll"));
            RegSetValueEx (hk, L"IsrDll", 0, REG_SZ, (BYTE *)L"\\Hard Disk\\AIO_COM\\israccesio.dll", sizeof(L"\\Hard Disk\\AIO_COM\\israccesio.dll"));
            RegSetValueEx (hk, L"IsrHandler", 0, REG_SZ, (BYTE *)L"ISRHandler", sizeof(L"ISRHandler"));
			RegSetValueEx (hk, L"ioctl", 0, REG_DWORD, (BYTE*)&dwIoctl, sizeof(DWORD));

                                        //@@@ Debug   Forcing devarrayIndex to 0
            dwIndex = 0;
            RegSetValueEx (hk, L"DeviceArrayIndex", 0, REG_DWORD, (BYTE*)&dwIndex, sizeof(DWORD));
			dwDeviceArrayIndex++; // incrementing here because index starts with 1
            dwIndex = dwDeviceArrayIndex + 3;
			RegSetValueEx (hk, L"Prefix", 0, REG_SZ, (BYTE *)L"COM", sizeof(L"COM"));
			RegSetValueEx (hk, L"Index", 0, REG_DWORD, (BYTE*)&dwIndex, sizeof(DWORD));
			dwPortInstanceOnChip = i;
			RegSetValueEx (hk, L"PortInstanceOnChip", 0, REG_DWORD, (BYTE*)&dwPortInstanceOnChip, sizeof(DWORD));
			RegSetValueEx (hk, L"DeviceID", 0, REG_DWORD, (BYTE*)&pHWHead->dwDeviceId, sizeof(DWORD));
			RegCloseKey (hk);

			// unimodem
            //wsprintf (szLine, L"Drivers\\BuiltIn\\PCI\\Template\\xrserial%d%x%d\\Unimodem", dwCardInstance, pHWHead->dwDeviceId, i);
            wsprintf (szLine, L"Drivers\\xrserial%d%x%d\\Unimodem", dwCardInstance, pHWHead->dwDeviceId, i);
			if (ERROR_SUCCESS == RegCreateKeyEx (HKEY_LOCAL_MACHINE, szLine, 0, NULL, 0, KEY_ALL_ACCESS, NULL, &hk, &dwDisp))
			{
				dwDeviceType = 0;
				RegSetValueEx (hk, L"DevConfig", 0, REG_BINARY, (BYTE *)ucDevConfig, sizeof(ucDevConfig));
				RegSetValueEx (hk, L"FriendlyName", 0, REG_SZ, (BYTE *)L"XR PCI COM Port", sizeof(L"XR PCI COM Port"));
				RegSetValueEx (hk, L"DeviceType", 0, REG_DWORD, (BYTE *)&dwDeviceType, sizeof(DWORD));
				RegSetValueEx (hk, L"Tsp", 0, REG_SZ, (BYTE *)L"Unimodem.dll", sizeof(L"Unimodem.dll"));
				RegCloseKey (hk);
			}

			// come back to the root registry; needed for ActivateDeviceEx
            //wsprintf (szLine, L"Drivers\\BuiltIn\\PCI\\Template\\xrserial%d%x%d", dwCardInstance, pHWHead->dwDeviceId, i);
            wsprintf (szLine, L"Drivers\\xrserial%d%x%d", dwCardInstance, pHWHead->dwDeviceId, i);
		}
		else
		{
            RETAILMSG( 1, (TEXT("Failed to set registry values; \r\n")));
		}


		//
		// Everything is set for the port driver, now Activate(Register) it.
		//
        RETAILMSG( 1, (TEXT("Calling ActivateDeviceEx [%s] \r\n"), szLine ));
		pHWHead->hDevice[i] = ActivateDeviceEx(szLine, 0, 0, 0);
		if (pHWHead->hDevice[i])
		{
            RETAILMSG( 1, (TEXT("ActivateDeviceEx success !!! \r\n")));
		}
		else
		{
            RETAILMSG( 1, (TEXT("ActivateDeviceEx failure !!! \r\n")));
		}
	}

	pHWHead->dwTotalPortsActivated = i;

    RETAILMSG( 1, (_T("XRP_Init: returning 0x%08x\r\n"), pHWHead));
	return ((DWORD)pHWHead); // return pHWHead. as this can be used in DeInit for Deactivating the port drivers
}


/*
 @doc EXTERNAL
 @func  BOOL | XRP_Deinit | De-initialize serial card.
 @parm DWORD | pHWHead | Context pointer returned from XRP_Init
 *
 @rdesc None.
 */
BOOL
XRP_Deinit(
	DWORD dwContext
	)
{
	PXRPCI_INFO pHWHead = (PXRPCI_INFO)dwContext;
	DWORD i;

	if ( !pHWHead )
	{
		/* Can't do much without this */
        RETAILMSG( 1, (TEXT("XRP_Deinit can't find pHWHead\r\n")));
		SetLastError(ERROR_INVALID_HANDLE);
		return(FALSE);
	}

	if (pHWHead->dwTotalPortsActivated)
	{
		for(i = 0; i < pHWHead->dwTotalPortsActivated; i++)
		{
			DeactivateDevice(pHWHead->hDevice[i]);
		}
	}

	LocalFree(pHWHead);
	return TRUE;
}


BOOL
XRP_IOControl(
	DWORD dwContext,
	DWORD Ioctl,
	PUCHAR pInBuf,
	DWORD InBufLen,
	PUCHAR pOutBuf,
	DWORD OutBufLen,
	PDWORD pdwBytesTransferred
)
{
	//
	// we don't support any IOCTLs at this point
	// BTW, is there any need for those? I don't think so
	//
	BOOL bRc = FALSE;
	return bRc;
}

VOID
XRP_PowerDown(
	DWORD dwContext
	)
{
	UNREFERENCED_PARAMETER(dwContext);
}


VOID
XRP_PowerUp(
	DWORD dwContext
)
{
	UNREFERENCED_PARAMETER(dwContext);
}


DWORD
XRP_Open(
	DWORD Context,
	DWORD Access,
	DWORD ShareMode)
{

    RETAILMSG( 1,(_T("XRP_Open(%x, 0x%x, 0x%x)\r\n"),Context, Access, ShareMode));

	UNREFERENCED_PARAMETER(Access);
	UNREFERENCED_PARAMETER(ShareMode);

	// pass back the device handle
	return Context;     // 0 indicates failure
}


BOOL
XRP_Close(
	DWORD Context
	)
{
    RETAILMSG( 1,(_T("XRP_Close(%x)\r\n"), Context));

	return TRUE;
}


DWORD
XRP_Read(
	DWORD dwContext,
	LPVOID pBuf,
	DWORD Len
	)
{
	UNREFERENCED_PARAMETER(dwContext);
	UNREFERENCED_PARAMETER(pBuf);
	UNREFERENCED_PARAMETER(Len);

    RETAILMSG( 1,(_T("XRP_Read\r\n")));
	SetLastError(ERROR_INVALID_FUNCTION);
	return 0;
}

DWORD
XRP_Write(
	DWORD dwContext,
	LPVOID pBuf,
	DWORD Len
	)
{
	UNREFERENCED_PARAMETER(dwContext);
	UNREFERENCED_PARAMETER(pBuf);
	UNREFERENCED_PARAMETER(Len);

    RETAILMSG( 1,(_T("XRP_Read\r\n")));
	SetLastError(ERROR_INVALID_FUNCTION);
	return 0;
}

ULONG
XRP_Seek(
	PVOID Context,
	LONG Position,
	DWORD Type
)
{
	UNREFERENCED_PARAMETER(Context);
	UNREFERENCED_PARAMETER(Position);
	UNREFERENCED_PARAMETER(Type);

	return (DWORD)-1;
}

BOOL
DllEntry(
	HANDLE hDllHandle,
	DWORD dwReason,
	LPVOID lpreserved
)
{
	BOOL bRc = TRUE;

	UNREFERENCED_PARAMETER(hDllHandle);
	UNREFERENCED_PARAMETER(lpreserved);

	switch (dwReason) {
		case DLL_PROCESS_ATTACH:
			{
			DEBUGREGISTER((HINSTANCE)hDllHandle);
            RETAILMSG( 1,(_T("*** DLL_PROCESS_ATTACH - XRPCI- Current Process: 0x%x, ID: 0x%x ***\r\n"),
				GetCurrentProcess(), GetCurrentProcessId()));
			}
			break;

		case DLL_PROCESS_DETACH:
			{
            RETAILMSG( 1,(_T("*** DLL_PROCESS_DETACH - XRPCI - Current Process: 0x%x, ID: 0x%x ***\r\n"),
				GetCurrentProcess(), GetCurrentProcessId()));
			}
			break;

		default:
			break;
	}

	return bRc;
}


/*++
*******************************************************************************
Routine:

	XRPCI_GetRegistryData

Description:

	Take the registry path provided to XRP_Init and use it to find this
	requested comm port's DeviceArrayIndex, the I/O base address, and the
	Interrupt number.

Arguments:

	LPCTSTR regKeyPath the registry path passed in to XRP_Init.

Return Value:

	-1 if there is an error.

*******************************************************************************
--*/
BOOL
XRPCI_GetRegistryData(PXRPCI_INFO pHWHead, LPCTSTR regKeyPath)
{
#define GCI_BUFFER_SIZE 256
	DWORD I;
	LONG regError;
	HKEY hKey;
	DWORD dwDataSize = GCI_BUFFER_SIZE;
	DDKISRINFO dii;
	DDKWINDOWINFO dwi;
	DDKPCIINFO ppi;

	// UART register stride (default = 1 byte).
	DWORD dwRegStride = 1;

    RETAILMSG( 1, (TEXT("Try to open %s\r\n"), regKeyPath));

	// We've been handed the name of a key in the registry that was generated
	// on the fly by device.exe. We're going to open that key and pull from it
	// a value that is the name of this serial port's real key. That key
	// will have the DeviceArrayIndex that we're trying to find.
	hKey = OpenDeviceKey(regKeyPath);
	if ( hKey == NULL ) {
        RETAILMSG( 1, (TEXT("Failed to open device key\r\n")));
		return ( FALSE );
	}

	// read interrupt configuration parameters
	dii.cbSize = sizeof(dii);
        RETAILMSG( 1, (TEXT("  GetIsrInfo... \r\n")));
	regError = DDKReg_GetIsrInfo(hKey, &dii);
	if(regError == ERROR_SUCCESS)
	{
        RETAILMSG( 1, (TEXT("  OK\r\n")));
		if(dii.dwSysintr != SYSINTR_NOP)
		{
			pHWHead->dwSysIntr = dii.dwSysintr;
		}
		else
		{
			regError = ERROR_FILE_NOT_FOUND;
		}

		pHWHead->dwIrq = dii.dwIrq;
	}

	//get our PCI device number
        RETAILMSG( 1, (TEXT("  GetPciInfo...\r\n ")));
	if ( regError == ERROR_SUCCESS )
	{
		ppi.cbSize = sizeof(ppi);
		regError = DDKReg_GetPciInfo(hKey, &ppi);
		if(regError == ERROR_SUCCESS)
		{
            RETAILMSG( 1,
				 (TEXT("PCI Device number - FOR debugging/understanding 0x%X, 0x%X\r\n"), ppi.dwDeviceNumber, ppi.idVals[PCIID_DEVICEID]));
			// to find out 158/154/152
			pHWHead->dwDeviceId = ppi.idVals[PCIID_DEVICEID];
		}
	}

	// get our device instance's index
	if(regError == ERROR_SUCCESS)
	{
        RETAILMSG( 1, (TEXT("  Devindex... \r\n")));
		dwDataSize = PC_REG_DEVINDEX_VAL_LEN;
		regError = RegQueryValueEx(
			hKey,
			PC_REG_DEVINDEX_VAL_NAME,
			NULL,
			NULL,
			(LPBYTE)(&pHWHead->dwDevIndex),
			&dwDataSize
		);
	}

	if ( regError == ERROR_SUCCESS )
	{
        RETAILMSG( 1, (TEXT("  OK,GetWindowInfo... \r\n")));
		dwi.cbSize = sizeof(dwi);
		regError = DDKReg_GetWindowInfo(hKey, &dwi);
		if(regError == ERROR_SUCCESS)
		{
        RETAILMSG( 1, (TEXT("  OK\r\n ")));
			if(dwi.dwNumIoWindows > 0)
			{
        RETAILMSG( 1, (TEXT("  dwNumIoWindows:%d \r\n"), dwi.dwNumIoWindows));
				pHWHead->dwInterfaceType = dwi.dwInterfaceType;
				pHWHead->dwBusNumber = dwi.dwBusNumber;

				pHWHead->dwIOBase = 0;
				for ( I = dwi.dwNumIoWindows-1; I >= 0; --I )
				{
					//The top 4 bits of the wrong base + 7 aren't writable, they're fixed at zero.
					//The top 4 bits of the right base + 7 are writable, they're part of the first scratchpad.
					//So if those bits are writable, then it's the right base.
					PUCHAR pScratchpadMaybe = (PUCHAR)(dwi.ioWindows[I].dwBase) + 7;
					UCHAR OriginalData, TestData;
					OriginalData = READ_PORT_UCHAR(pScratchpadMaybe);
					WRITE_PORT_UCHAR(pScratchpadMaybe, OriginalData | 0xF0);
					TestData = READ_PORT_UCHAR(pScratchpadMaybe);
					WRITE_PORT_UCHAR(pScratchpadMaybe, OriginalData);
					if ( (TestData & 0xF0) != 0 )
					{
        RETAILMSG( 1, (TEXT("  Scratch pad found...\r\n ")));
						pHWHead->dwIOBase = dwi.ioWindows[I].dwBase;
						pHWHead->dwIOLen = dwi.ioWindows[I].dwLen;
        RETAILMSG( 1, (TEXT("  IOBase: %d, Len:%d\r\n "), pHWHead->dwIOBase, pHWHead->dwIOLen ));
						break;
					}
				}
                if ( pHWHead->dwIOBase == 0 )
				{
        RETAILMSG( 1, (TEXT("  ioBase not 0...\r\n ")));
					regError = ERROR_FILE_NOT_FOUND;
				}
			}
			else
			{
        RETAILMSG( 1, (TEXT("  invalid iowindows...\r\n ")));
				regError = ERROR_FILE_NOT_FOUND;
			}
		}
	}


	// get device register stride.
	if ( regError == ERROR_SUCCESS )
	{
        RETAILMSG( 1, (TEXT("  REGSTRIDE... ")));
		dwDataSize = PC_REG_REGSTRIDE_VAL_LEN;
		RegQueryValueEx(hKey,
				PC_REG_REGSTRIDE_VAL_NAME,
				NULL,
				NULL,
				(LPBYTE)(&dwRegStride),
				&dwDataSize);

        RETAILMSG( 1,
				 (TEXT("dwRegStride - FOR debugging/understanding 0x%X\r\n"), dwRegStride));
	}

	RegCloseKey (hKey);

	if ( regError != ERROR_SUCCESS )
	{
        RETAILMSG( 1,
				 (TEXT("Failed to get XR PCI registry values, Error 0x%X\r\n"), regError));
		return ( FALSE );
	}

    RETAILMSG( 1,
		(TEXT("XRPCI_GetRegistryData - Devindex %d, SysIntr %d, IOB %X, IOLen %X \r\n"),
		pHWHead->dwDevIndex, pHWHead->dwSysIntr, pHWHead->dwIOBase, pHWHead->dwIOLen)
	);

	return ( TRUE );
}
